import os
import subprocess
import re
from glob import glob
from pathlib import Path
from typing import Callable, Tuple

from cr8.misc import parse_version


MIN_VERSION_FOR_JVM11 = (2, 3, 6)
VERSION_RE = re.compile(r'(\d+\.\d+\.\d+(_\d+)?)|"(\d+)"')


def _get_candidates() -> Tuple[str, ...]:
    return tuple(
        glob("/usr/lib/jvm/java-*-openjdk")
        + glob("/usr/lib/java-*")
        + glob("/Library/Java/JavaVirtualMachines/jdk*/Contents/Home")
        + glob("/Library/Java/JavaVirtualMachines/*/Contents/Home")
        + glob(os.path.expanduser("~/.m2/jdks/jdk-*"))
    )


def _parse_java_version(line: str) -> tuple:
    """ Return the version number found in the first line of `java -version`

    >>> _parse_java_version('openjdk version "11.0.2" 2018-10-16')
    (11, 0, 2)
    """
    m = VERSION_RE.search(line)
    version_str = m and m.group(0).replace('"', '') or '0.0.0'
    if '_' in version_str:
        fst, snd = version_str.split('_', maxsplit=2)
        version = parse_version(fst)
        return (version[1], version[2], int(snd))
    else:
        return parse_version(version_str)


def _detect_java_version(java_home: str) -> tuple:
    p = subprocess.run(
        [Path(java_home) / 'bin' / 'java', '-version'],
        check=True,
        encoding='utf-8',
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT
    )
    line = p.stdout.split('\n')[0]
    return _parse_java_version(line)


def _find_matching_java_home(version_matches: Callable[[tuple], bool]) -> str:
    java_home = os.environ.get('JAVA_HOME', '')
    for path in filter(os.path.exists, (java_home, ) + _get_candidates()):
        version = _detect_java_version(path)
        if version_matches(version):
            return path
    return java_home


def find_java_home(cratedb_version: tuple) -> str:
    """ Return a path to a JAVA_HOME suites for the given CrateDB version """
    if MIN_VERSION_FOR_JVM11 <= cratedb_version < (4, 0):
        return _find_matching_java_home(lambda ver: ver[0] <= 11)
    if cratedb_version < MIN_VERSION_FOR_JVM11:
        return _find_matching_java_home(lambda ver: ver[0] == 8)
    if cratedb_version <= (4, 1, 0):
        return _find_matching_java_home(lambda ver: ver[0] == 11)
    return _find_matching_java_home(lambda ver: ver[0] >= 11)




import tempfile
import shutil
from unittest import IsolatedAsyncioTestCase
from cr8.run_crate import CrateNode, get_crate
from cr8.reindex import areindex
from cr8.clients import client


class TestReindex(IsolatedAsyncioTestCase):

    def setUp(self):
        self._to_stop = []
        self.data_path = tempfile.mkdtemp()
        self.crate_settings = {
            'path.data': self.data_path,
            'cluster.name': 'cr8-reindex-tests',
            'http.port': '44200-44250'
        }

    def tearDown(self):
        for node in self._to_stop:
            node.stop()
        self._to_stop = []
        shutil.rmtree(self.data_path, ignore_errors=True)

    async def test_reindex(self):
        crate_v5 = CrateNode(
            crate_dir=get_crate('5.x.x'),
            keep_data=True,
            settings=self.crate_settings
        )
        self._to_stop.append(crate_v5)
        crate_v5.start()
        async with client(crate_v5.http_url) as c:
            await c.execute("create table t (x int)")
            args = (
                (1,),
                (2,),
                (3,),
            )
            await c.execute_many("insert into t (x) values (?)", args)
        crate_v5.stop()
        self._to_stop.remove(crate_v5)

        crate_v6 = CrateNode(
            crate_dir=get_crate('6.0.0'),
            keep_data=True,
            settings=self.crate_settings
        )
        self._to_stop.append(crate_v6)
        crate_v6.start()
        await areindex(hosts=crate_v6.http_url)
        async with client(crate_v6.http_url) as c:
            result = await c.execute("SELECT version FROM information_schema.tables WHERE table_name = 't'")
            version = result['rows'][0][0]
            self.assertEqual(version, {'upgraded': None, 'created': '6.0.0'})

            result = await c.execute('SELECT count(*) FROM t')
            cnt = result['rows'][0][0]
            self.assertEqual(cnt, 3)



import json

import aiohttp
import itertools
import calendar
import types
import time
import contextlib
from urllib.parse import urlparse, parse_qs, urlunparse
from datetime import datetime, date
from typing import List, Union, Iterable, Dict, Optional, Any
from decimal import Decimal
from cr8.aio import asyncio  # import via aio for uvloop setup

try:
    import asyncpg
except ImportError:
    asyncpg = None  # type: ignore

try:
    import simdjson  # type: ignore
    dumps = simdjson.dumps
except ImportError:
    dumps = json.dumps


HTTP_DEFAULT_HDRS = {'Content-Type': 'application/json'}


class CrateJsonEncoder(json.JSONEncoder):

    epoch = datetime(1970, 1, 1)

    def default(self, o):
        if isinstance(o, Decimal):
            return str(o)
        if isinstance(o, datetime):
            delta = o.replace(tzinfo=None) - self.epoch
            return int(delta.total_seconds() * 1000)
        if isinstance(o, date):
            return calendar.timegm(o.timetuple()) * 1000
        return json.JSONEncoder.default(self, o)


class SqlException(Exception):

    def __init__(self, message):
        self.message = message


client_errors = [
    SqlException,
    aiohttp.ClientError
]
if asyncpg:
    client_errors.append(asyncpg.exceptions.PostgresError)


def _to_http_uri(s: str) -> str:
    """Prefix the string with 'http://' if there is no schema."""
    if not s.startswith(('http://', 'https://')):
        return 'http://' + s
    return s


def _to_http_hosts(hosts: Union[Iterable[str], str]) -> List[str]:
    """Convert a string of whitespace or comma separated hosts into a list of hosts.

    Hosts may also already be a list or other iterable.
    Each host will be prefixed with 'http://' if it is not already there.

    >>> _to_http_hosts('n1:4200,n2:4200')
    ['http://n1:4200', 'http://n2:4200']

    >>> _to_http_hosts('n1:4200 n2:4200')
    ['http://n1:4200', 'http://n2:4200']

    >>> _to_http_hosts('https://n1:4200')
    ['https://n1:4200']

    >>> _to_http_hosts('https://n1:4200/?verify_ssl=false')
    ['https://n1:4200/?verify_ssl=false']

    >>> _to_http_hosts(['http://n1:4200', 'n2:4200'])
    ['http://n1:4200', 'http://n2:4200']
    """
    if isinstance(hosts, str):
        hosts = hosts.replace(',', ' ').split()
    return [_to_http_uri(i) for i in hosts]


async def _exec(session, url, data):
    async with session.post(url,
                            data=data,
                            headers=HTTP_DEFAULT_HDRS,
                            timeout=None) as resp:
        if resp.status == 401:
            t = await resp.text()
            raise SqlException(t)
        r = await resp.json()
        if 'error' in r:
            raise SqlException(
                r['error']['message'] + ' occurred using: ' + str(data))
        return r


def _plain_or_callable(obj):
    """Returns the value of the called object of obj is a callable,
    otherwise the plain object.
    Returns None if obj is None.

    >>> obj = None
    >>> _plain_or_callable(obj)

    >>> stmt = 'select * from sys.nodes'
    >>> _plain_or_callable(stmt)
    'select * from sys.nodes'

    >>> def _args():
    ...     return [1, 'name']
    >>> _plain_or_callable(_args)
    [1, 'name']

    >>> _plain_or_callable((x for x in range(10)))
    0

    >>> class BulkArgsGenerator:
    ...     def __call__(self):
    ...         return [[1, 'foo'], [2, 'bar'], [3, 'foobar']]
    >>> _plain_or_callable(BulkArgsGenerator())
    [[1, 'foo'], [2, 'bar'], [3, 'foobar']]
    """
    if callable(obj):
        return obj()
    elif isinstance(obj, types.GeneratorType):
        return next(obj)
    else:
        return obj


def _date_or_none(d: str) -> Optional[str]:
    """Return a date as if, if valid, otherwise None

    >>> _date_or_none('2017-02-27')
    '2017-02-27'

    >>> _date_or_none('NA')
    """
    try:
        datetime.strptime(d, '%Y-%m-%d')
        return d
    except ValueError:
        return None


def _to_dsn(hosts: str) -> str:
    """Convert a host URI into a dsn for aiopg.

    >>> _to_dsn('aiopg://myhostname:4242/mydb')
    'postgres://crate@myhostname:4242/mydb'

    >>> _to_dsn('aiopg://myhostname:4242')
    'postgres://crate@myhostname:4242/doc'

    >>> _to_dsn('aiopg://hoschi:pw@myhostname:4242/doc?sslmode=require')
    'postgres://hoschi:pw@myhostname:4242/doc?sslmode=require'

    >>> _to_dsn('aiopg://myhostname')
    'postgres://crate@myhostname:5432/doc'
    """
    p = urlparse(hosts)
    try:
        user_and_pw, netloc = p.netloc.split('@', maxsplit=1)
    except ValueError:
        netloc = p.netloc
        user_and_pw = 'crate'
    try:
        host, port = netloc.split(':', maxsplit=1)
    except ValueError:
        host = netloc
        port = "5432"
    dbname = p.path[1:] if p.path else 'doc'
    dsn = f'postgres://{user_and_pw}@{host}:{port}/{dbname}'
    if p.query:
        dsn += '?' + '&'.join(k + '=' + v[0] for k, v in parse_qs(p.query).items())
    return dsn


def _to_boolean(v: str) -> bool:
    if str(v).lower() in ("true"):
        return True
    elif str(v).lower() in ("false"):
        return False
    else:
        raise ValueError('not a boolean value')


def _verify_ssl_from_first(hosts: List[str]) -> bool:
    """Check if SSL validation parameter is passed in URI

    >>> _verify_ssl_from_first(['https://myhost:4200/?verify_ssl=false'])
    False

    >>> _verify_ssl_from_first(['https://myhost:4200/'])
    True

    >>> _verify_ssl_from_first([
    ...      'https://h1:4200/?verify_ssl=False',
    ...      'https://h2:4200/?verify_ssl=True'
    ... ])
    False
    """
    for host in hosts:
        query = parse_qs(urlparse(host).query)
        if 'verify_ssl' in query:
            return _to_boolean(query['verify_ssl'][0])
    return True


class AsyncpgClient:
    def __init__(self, hosts, pool_size=25, session_settings=None):
        self.dsn = _to_dsn(hosts)
        self.pool_size = pool_size
        self._pool = None
        self.is_cratedb = True
        self.session_settings = session_settings or {}

    async def _get_pool(self):

        async def set_session_settings(conn):
            for setting, value in self.session_settings.items():
                await conn.execute(f'set {setting}={value}')

        if not self._pool:
            self._pool = await asyncpg.create_pool(
                self.dsn,
                min_size=self.pool_size,
                max_size=self.pool_size,
                init=set_session_settings
            )
        return self._pool

    async def execute(self, stmt, args=None):
        start = time.perf_counter()
        pool = await self._get_pool()
        async with pool.acquire() as conn:
            if args:
                rows = await conn.fetch(stmt, *args)
            else:
                rows = await conn.fetch(stmt)
            return {
                'duration': (time.perf_counter() - start) * 1000.,
                'rows': rows
            }

    async def execute_many(self, stmt, bulk_args):
        start = time.perf_counter()
        pool = await self._get_pool()
        async with pool.acquire() as conn:
            await conn.executemany(stmt, bulk_args)
            return {
                'duration': (time.perf_counter() - start) * 1000.,
                'rows': []
            }

    async def get_server_version(self):
        pool = await self._get_pool()
        async with pool.acquire() as conn:
            try:
                for (version,) in await conn.fetch('select version from sys.nodes'):
                    version = json.loads(version)
                    return {
                        'hash': version['build_hash'],
                        'number': version['number']
                    }
            except asyncpg.exceptions.UndefinedTableError:
                self.is_cratedb = False
                version = await conn.fetchval("select current_setting('server_version_num')")
                return {
                    'hash': None,
                    'number': version
                }

    async def close(self):
        if self._pool:
            await self._pool.close()
            self._pool = None

    async def __aenter__(self):
        return self

    async def __aexit__(self, *exs):
        await self.close()


def _append_sql(host: str) -> str:
    """ Append `/_sql` to the host, dropping any query parameters.

    >>> _append_sql('http://n1:4200')
    'http://n1:4200/_sql'

    >>> _append_sql('https://n1:4200/?verify_ssl=false')
    'https://n1:4200/_sql'

    >>> _append_sql('https://crate@n1:4200/?verify_ssl=false')
    'https://crate@n1:4200/_sql'

    """
    p = list(urlparse(host))
    p[2] = '_sql'
    p[3] = None
    p[4] = None
    return urlunparse(tuple(p))


class HttpClient:
    def __init__(self,
                 hosts: List[str],
                 conn_pool_limit=25,
                 session_settings: Optional[Dict[str, Any]] = None):
        self.hosts = hosts
        self.urls = itertools.cycle(list(map(_append_sql, hosts)))
        self.conn_pool_limit = conn_pool_limit
        self.is_cratedb = True
        self._pools: Dict[str, asyncio.Queue] = {}
        self.session_settings = session_settings or {}

    @contextlib.asynccontextmanager
    async def _session(self, url):
        pool = self._pools.get(url)
        if not pool:
            pool = asyncio.Queue(maxsize=self.conn_pool_limit)
            self._pools[url] = pool
            _connector_params = {
                'limit': 1,
                'verify_ssl': _verify_ssl_from_first(self.hosts)
            }
            for n in range(0, self.conn_pool_limit):
                tcp_connector = aiohttp.TCPConnector(**_connector_params)
                session = aiohttp.ClientSession(connector=tcp_connector)
                for setting, value in self.session_settings.items():
                    payload = {'stmt': f'set {setting}={value}'}
                    await _exec(
                        session,
                        url,
                        dumps(payload, cls=CrateJsonEncoder)
                    )
                pool.put_nowait(session)

        session = await pool.get()
        try:
            yield session
        finally:
            await pool.put(session)

    async def execute(self, stmt, args=None):
        payload = {'stmt': _plain_or_callable(stmt)}
        if args:
            payload['args'] = _plain_or_callable(args)
        url = next(self.urls)
        async with self._session(url) as session:
            result = await _exec(
                session,
                url,
                dumps(payload, cls=CrateJsonEncoder)
            )
        return result

    async def execute_many(self, stmt, bulk_args):
        data = dumps(dict(
            stmt=_plain_or_callable(stmt),
            bulk_args=_plain_or_callable(bulk_args)
        ), cls=CrateJsonEncoder)
        url = next(self.urls)
        async with self._session(url) as session:
            result = await _exec(session, url, data)
        return result

    async def get_server_version(self):
        urlparts = urlparse(self.hosts[0])
        url = urlunparse((urlparts.scheme, urlparts.netloc, '/', '', '', ''))
        async with self._session(url) as session:
            async with session.get(url) as resp:
                r = await resp.json()
                version = r['version']
                result = {
                    'hash': version['build_hash'],
                    'number': version['number'],
                    'date': _date_or_none(version['build_timestamp'][:10])
                }
                return result

    async def close(self):
        pools = self._pools
        self._pools = {}
        for url, pool in pools.items():
            for i in range(0, self.conn_pool_limit):
                session = await pool.get()
                await session.close()
        pools.clear()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()


def client(hosts, session_settings=None, concurrency=25):
    hosts = hosts or 'localhost:4200'
    if hosts.startswith('asyncpg://'):
        if not asyncpg:
            raise ValueError('Cannot use "asyncpg" scheme if asyncpg is not available')
        return AsyncpgClient(hosts, pool_size=concurrency, session_settings=session_settings)
    return HttpClient(_to_http_hosts(hosts), conn_pool_limit=concurrency, session_settings=session_settings)



"""misc functions that have no real home."""

import logging
import gzip
from pathlib import Path
from urllib.request import urlopen
from typing import Tuple, Iterator, Any, Optional
from collections import defaultdict


def init_logging(log):
    log.setLevel(logging.INFO)
    ch = logging.StreamHandler()
    ch.setLevel(logging.INFO)
    log.addHandler(ch)


def try_len(o: Any) -> Optional[int]:
    """ Return len of `o` or None if `o` doesn't support len

    >>> try_len([1, 2])
    2

    >>> try_len(print)
    >>> try_len(None)
    """
    if not o:
        return None
    try:
        return len(o)
    except TypeError:
        return None


def parse_version(version: str) -> tuple[int, int, int]:
    """Parse a string formatted X[.Y.Z] version number into a tuple

    >>> parse_version('10.2.3')
    (10, 2, 3)

    >>> parse_version('12')
    (12, 0, 0)
    """
    parts = version.split('.')
    if len(parts) == 1:
        return (int(parts[0]), 0, 0)
    if len(parts) == 2:
        return (int(parts[0]), int(parts[1]), 0)
    return (int(parts[0]), int(parts[1]), int(parts[2]))


def parse_table(fq_table: str) -> Tuple[str, str]:
    """Parse a tablename into tuple(<schema>, <table>).

    Schema defaults to doc if the table name doesn't contain a schema.

    >>> parse_table('x.users')
    ('x', 'users')

    >>> parse_table('users')
    ('doc', 'users')
    """
    parts = fq_table.split('.')
    if len(parts) == 1:
        return 'doc', parts[0]
    elif len(parts) == 2:
        return parts[0], parts[1]
    else:
        raise ValueError


def as_bulk_queries(queries, bulk_size):
    """Group a iterable of (stmt, args) by stmt into (stmt, bulk_args).

    bulk_args will be a list of the args grouped by stmt.

    len(bulk_args) will be <= bulk_size
    """
    stmt_dict = defaultdict(list)
    for stmt, args in queries:
        bulk_args = stmt_dict[stmt]
        bulk_args.append(args)
        if len(bulk_args) == bulk_size:
            yield stmt, bulk_args
            del stmt_dict[stmt]
    for stmt, bulk_args in stmt_dict.items():
        yield stmt, bulk_args


def get_lines(filename: str) -> Iterator[str]:
    """Create an iterator that returns the lines of a utf-8 encoded file."""

    if filename.startswith(('https://', 'http://')):
        uri = filename
    else:
        path = Path(filename)
        if not path.is_absolute():
            path = Path.cwd() / path
        uri = path.as_uri()

    with urlopen(uri) as f:
        if filename.endswith('.gz'):
            with gzip.GzipFile(fileobj=f) as fz:
                for line in fz:
                    yield line.decode('utf-8')
        else:
            for line in f:
                yield line.decode('utf-8')


def as_statements(lines: Iterator[str]) -> Iterator[str]:
    """Create an iterator that transforms lines into sql statements.

    Statements within the lines must end with ";"
    The last statement will be included even if it does not end in ';'

    >>> list(as_statements(['select * from', '-- comments are filtered', 't;']))
    ['select * from t']

    >>> list(as_statements(['a;', 'b', 'c;', 'd', ' ']))
    ['a', 'b c', 'd']
    """
    lines = (l.strip() for l in lines if l)
    lines = (l for l in lines if l and not l.startswith('--'))
    parts = []
    for line in lines:
        parts.append(line.rstrip(';'))
        if line.endswith(';'):
            yield ' '.join(parts)
            parts.clear()
    if parts:
        yield ' '.join(parts)


def break_iterable(iterable, pred):
    """Break a iterable on the item that matches the predicate into lists.

    The item that matched the predicate is not included in the result.

    >>> list(break_iterable([1, 2, 3, 4], lambda x: x == 3))
    [[1, 2], [4]]
    """
    sublist = []
    for i in iterable:
        if pred(i):
            yield sublist
            sublist = []
        else:
            sublist.append(i)
    yield sublist



import os
import doctest
import subprocess
import functools
from unittest import TestCase

from cr8.run_crate import CrateNode, get_crate
from cr8.clients import client
from cr8 import aio


crate_dir = get_crate('latest-testing')
node = CrateNode(
    crate_dir=crate_dir,
    settings={
        'cluster.name': 'cr8-tests',
        'http.port': '44200-44250'
    })


def setup(*args):
    async def run():
        async with client(node.http_url) as c:
            await c.execute(
                'create table x.demo (id int, name string, country string) \
                with (number_of_replicas = 0)'
            )
            await c.execute('create table y.demo (name text) with (number_of_replicas = 0)')
            await c.execute('create blob table blobtable with (number_of_replicas = 0)')
    aio.run(run())


def teardown(*args):
    async def run():
        async with client(node.http_url) as c:
            await c.execute('drop table x.demo')
            await c.execute('drop table y.demo')
            await c.execute('drop blob table blobtable')
    try:
        aio.run(run())
    finally:
        node.stop()


def transform(s):
    s = s.replace('localhost:4200', node.http_url)
    s = s.replace(
        'asyncpg://localhost:5432',
        f'asyncpg://{node.addresses.psql.host}:{node.addresses.psql.port}')
    s = s.replace(
        'postgresql://crate@localhost:5432/doc',
        f'postgresql://crate@{node.addresses.psql.host}:{node.addresses.psql.port}/doc')
    return (
        r'print(sh("""%s""").stdout.decode("utf-8"))' % s) + '\n'


class Parser(doctest.DocTestParser):

    def parse(self, string, name='<string>'):
        r = super().parse(string, name)
        for s in r:
            if isinstance(s, doctest.Example):
                s.source = transform(s.source)
        return r


class SourceBuildTest(TestCase):

    def test_build_from_branch(self):
        self.assertIsNotNone(get_crate('5.8'))


def load_tests(loader, tests, ignore):
    env = os.environ.copy()
    env['CR8_NO_TQDM'] = 'True'
    node.start()
    assert node.http_host, "http_url must be available"
    tests.addTests(doctest.DocFileSuite(
        os.path.join('..', 'README.rst'),
        globs={
            'sh': functools.partial(
                subprocess.run,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                timeout=60,
                shell=True,
                env=env
            )
        },
        optionflags=doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS,
        setUp=setup,
        tearDown=teardown,
        parser=Parser()
    ))
    return tests


if __name__ == "__main__":
    import unittest
    unittest.main()



"""
reindex all tables that require re-indexing to upgrade to the next major CrateDB version.

This temporarily copies all the data, so additional disk space is required and
as it re-indexed the data the cluster will experience higher load while this is
running.
"""

import argh
from cr8 import clients, aio
from cr8.misc import parse_version


FIND_TABLES_4_0 = '''
SELECT
    table_schema,
    table_name
FROM
    information_schema.tables
WHERE
    substr(version['created'], 0, 1)::int < (
        SELECT
            max(substr(version['number'], 0, 1)::int)
        FROM
            sys.nodes);
'''

FIND_TABLES_3_3 = '''
SELECT
    table_schema,
    table_name
FROM
    information_schema.tables
WHERE
    substr(version['created']['cratedb'], 0, 1)::int < (
        SELECT
            max(substr(version['number'], 0, 1)::int)
        FROM
            sys.nodes);
'''


COL_NAMES = '''
SELECT
    column_name
FROM
    information_schema.columns
WHERE
    table_schema = ?
    AND table_name = ?
    AND CAST(is_generated AS string) in ('f', 'NEVER')
    AND column_name NOT LIKE '%]'
ORDER BY
    ordinal_position ASC,
    column_name;
'''


async def _fetch_tables_to_upgrade(client, version):
    query = FIND_TABLES_3_3 if version < (4, 0, 0) else FIND_TABLES_4_0
    result = await client.execute(query)
    return result['rows']


async def _show_create_table(client, fq_table):
    result = await client.execute(f'SHOW CREATE TABLE {fq_table}')
    return result['rows'][0][0]


async def _reindex_table(client, schema, table, create_table, column_names):
    tmp_table = f'tmp__{table}'
    print(f'Creating {schema}.{tmp_table}')
    create_tmp_table = create_table.replace(
        f'"{schema}"."{table}"',
        f'"{schema}"."{tmp_table}"'
    )
    await client.execute(create_tmp_table)
    columns = ', '.join((f'"{x}"' for x in column_names))
    print(f'Copy data from "{schema}"."{table}" to "{schema}"."{tmp_table}"')
    await client.execute(
        f'INSERT INTO "{schema}"."{tmp_table}" ({columns}) (SELECT {columns} FROM "{schema}"."{table}")')
    print(f'Replacing {table} with {tmp_table}')
    await client.execute(
        f'ALTER CLUSTER SWAP TABLE "{schema}"."{tmp_table}" TO "{schema}"."{table}" WITH (drop_source = true)')


async def _fetch_column_names(client, schema, table):
    result = await client.execute(COL_NAMES, (schema, table))
    return (row[0] for row in result['rows'])


async def _async_reindex(client):
    version_dict = await client.get_server_version()
    version = parse_version(version_dict['number'])
    if version < (3, 3, 0):
        raise ValueError("reindex only works on a CrateDB cluster running 3.3.0 or later")
    tables = await _fetch_tables_to_upgrade(client, version)
    for schema, table in tables:
        if schema == 'blob':
            print(f'Skipping unsupported BLOB table "{schema}"."{table}"')
            continue
        create_table = await _show_create_table(client, f'"{schema}"."{table}"')
        column_names = await _fetch_column_names(client, schema, table)
        await _reindex_table(client, schema, table, create_table, column_names)


async def areindex(*, hosts=None):
    async with clients.client(hosts) as client:
        await _async_reindex(client)


@argh.arg('--hosts', help='crate hosts', type=str, required=True)
def reindex(*, hosts=None):
    async def run():
        await areindex(hosts=hosts)
    aio.run(run())


def main():
    argh.dispatch_command(reindex)


reindex.__doc__ = __doc__




import argh
import asyncpg
import asyncio

from functools import partial

from cr8 import aio, clients
from cr8.cli import to_int
from cr8.metrics import Stats
from cr8.log import format_stats


def mk_insert(table, attributes):
    columns = ', '.join(('"' + x.name + '"' for x in attributes))
    params = ', '.join((f'${i + 1}' for i in range(len(attributes))))
    return f'INSERT INTO {table} ({columns}) VALUES ({params})'


async def reader(q, conn, stmt, fetch_size):
    async with conn.transaction():
        bulk_args = []
        async for row in stmt.cursor(prefetch=fetch_size):
            bulk_args.append(tuple(row))
            if len(bulk_args) == fetch_size:
                await q.put(bulk_args)
                bulk_args = []
        if bulk_args:
            await q.put(bulk_args)
        await q.put(None)


async def writer(q, insert, insert_many):
    with aio.tqdm() as t:
        while True:
            bulk_args = await q.get()
            if bulk_args is None:
                break
            await insert_many(insert, bulk_args)
            t.update(1)


async def async_insert_from_sql(src_uri,
                                concurrency,
                                query,
                                fetch_size,
                                table,
                                insert_many):
    conn = await asyncpg.connect(src_uri)
    try:
        stmt = await conn.prepare(query)
        insert = mk_insert(table, stmt.get_attributes())
        print(insert)
        q = asyncio.Queue(maxsize=concurrency)
        await asyncio.gather(
            reader(q, conn, stmt, fetch_size),
            writer(q, insert, insert_many)
        )
    finally:
        await conn.close()


@argh.arg('--src-uri', help='source uri', required=True)
@argh.arg('--query', help='select statement', required=True)
@argh.arg('--fetch-size', type=to_int)
@argh.arg('--table', help='Target table', required=True)
@argh.arg('--hosts', help='Target CrateDB hosts')
@argh.arg('-c', '--concurrency', type=to_int)
@argh.arg('-of', '--output-fmt', choices=['json', 'text'], default='text')
@argh.wrap_errors([KeyboardInterrupt, BrokenPipeError] + clients.client_errors)
def insert_from_sql(*,
                    src_uri=None,
                    query=None,
                    fetch_size=100,
                    concurrency=25,
                    table=None,
                    hosts=None,
                    output_fmt=None):
    """Insert data read from another SQL source into table."""

    stats = Stats()

    async def run():
        async with clients.client(hosts, concurrency=concurrency) as client:
            f = partial(aio.measure, stats, client.execute_many)
            try:
                await async_insert_from_sql(
                    src_uri,
                    concurrency,
                    query,
                    fetch_size,
                    table,
                    f
                )
            except clients.SqlException as e:
                raise SystemExit(str(e))
        try:
            print(format_stats(stats.get(), output_fmt))
        except KeyError:
            if not stats.sampler.values:
                raise SystemExit('No data read from source')
            raise
    aio.run(run())


def main():
    argh.dispatch_command(insert_from_sql)


if __name__ == "__main__":
    main()



import os
from unittest import TestCase
from doctest import DocTestSuite

from cr8.bench_spec import load_spec

from cr8 import engine


class SpecTest(TestCase):

    def test_session_settings_from_spec(self):
        spec = self.get_spec('sample.py')
        self.assertEqual(spec.session_settings, {'application_name': 'my_app', 'timezone': 'UTC'})

    def test_session_settings_from_toml(self):
        spec = self.get_spec('sample.toml')
        self.assertEqual(spec.session_settings, {'application_name': 'my_app', 'timezone': 'UTC'})

    def test_session_settings_from_json(self):
        spec = self.get_spec('count_countries.json')
        self.assertEqual(spec.session_settings, {'application_name': 'my_app', 'timezone': 'UTC'})

    def get_spec(self, name):
        return load_spec(os.path.abspath(os.path.join(os.path.dirname(__file__), '../specs/', name)))


def load_tests(loader, tests, ignore):
    tests.addTests(DocTestSuite(engine))
    return tests



try:
    from importlib.metadata import version, PackageNotFoundError
    __version__ = version("cr8")
except ImportError:
    try:
        import pkg_resources
        __version__ = pkg_resources.require('cr8')[0].version
    except ImportError:
        __version__ = "unknown"
except PackageNotFoundError:
    __version__ = "unknown"



from unittest import TestCase
from doctest import DocTestSuite
from cr8 import java_magic
from cr8.java_magic import _parse_java_version


class JavaVersionParsingTest(TestCase):

    def assertVersion(self, line, expected):
        version = _parse_java_version(line)
        self.assertEqual(version, expected)

    def test_java_8_line(self):
        self.assertVersion('openjdk version "1.8.0_202"', (8, 0, 202))

    def test_java_10_line(self):
        self.assertVersion('openjdk version "10.0.2" 2018-07-17', (10, 0, 2))

    def test_java_11_line(self):
        self.assertVersion('java 11.0.1 2018-10-16 LTS', (11, 0, 1))

    def test_java_12_line(self):
        self.assertVersion('openjdk version "12" 2019-03-19', (12, 0, 0))


def load_tests(loader, tests, ignore):
    tests.addTests(DocTestSuite(java_magic))
    return tests
